# hobo - portal to configure and deploy applications
# Copyright (C) 2015-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import json
import os
import sys
import tarfile
import urllib.parse

from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.core.files.base import ContentFile
from django.db import connection, models
from django.utils.text import slugify
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from hobo.environment.utils import get_installed_services

from .utils import Requests

requests = Requests()


class Application(models.Model):
    SUPPORTED_MODULES = ('wcs',)

    name = models.CharField(max_length=100, verbose_name=_('Name'))
    slug = models.SlugField(max_length=100)
    icon = models.FileField(
        verbose_name=_('Icon'),
        help_text=_(
            'Icon file must be in JPEG or PNG format, and should be a square of at least 512×512 pixels.'
        ),
        upload_to='applications/icons/',
        blank=True,
        null=True,
    )
    description = models.TextField(verbose_name=_('Description'), blank=True)
    documentation_url = models.URLField(_('Documentation URL'), blank=True)
    editable = models.BooleanField(default=True)
    elements = models.ManyToManyField('Element', blank=True, through='Relation')
    creation_timestamp = models.DateTimeField(default=now)
    last_update_timestamp = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return '<Application %s>' % self.slug

    def save(self, *args, **kwargs):
        if not self.slug:
            base_slug = slugify(self.name)[:95]
            slug = base_slug
            i = 1
            while Application.objects.filter(slug=slug).exists():
                slug = '%s-%s' % (base_slug, i)
                i += 1
            self.slug = slug
        super().save(*args, **kwargs)

    def scandeps(self):
        self.relation_set.filter(auto_dependency=True).delete()
        relations = self.relation_set.select_related('element')
        elements = {(x.element.type, x.element.slug): (x.element, x) for x in relations}
        finished = False
        while not finished:
            finished = True
            for el in [x[0] for x in elements.values()]:
                dependencies_url = el.cache['urls'].get('dependencies')
                if not dependencies_url:
                    continue
                response = requests.get(dependencies_url)
                for dependency in response.json()['data']:
                    if (dependency['type'], dependency['id']) in elements:
                        continue
                    finished = False
                    element, created = Element.objects.get_or_create(
                        type=dependency['type'], slug=dependency['id'], defaults={'name': dependency['text']}
                    )
                    element.name = dependency['text']
                    element.cache = dependency
                    element.save()
                    relation, created = Relation.objects.get_or_create(application=self, element=element)
                    if created:
                        relation.auto_dependency = True
                        relation.save()
                    elements[(element.type, element.slug)] = (element, relation)
        return elements


class Element(models.Model):
    type = models.CharField(max_length=25, verbose_name=_('Type'))
    slug = models.SlugField(max_length=500, verbose_name=_('Slug'))
    name = models.CharField(max_length=500, verbose_name=_('Name'))
    cache = JSONField(blank=True, default=dict)

    def __repr__(self):
        return '<Element %s/%s>' % (self.type, self.slug)

    def get_redirect_url(self):
        if self.type == 'roles':
            return
        if not self.cache.get('urls'):
            return
        if self.cache['urls'].get('redirect'):
            return self.cache['urls']['redirect']
        if self.cache['urls'].get('export'):
            return '%sredirect/' % self.cache['urls']['export']


class Relation(models.Model):
    application = models.ForeignKey(Application, on_delete=models.CASCADE)
    element = models.ForeignKey(Element, on_delete=models.CASCADE)
    auto_dependency = models.BooleanField(default=False)

    def __repr__(self):
        return '<Relation %s - %s/%s>' % (self.application.slug, self.element.type, self.element.slug)


class Version(models.Model):
    application = models.ForeignKey(Application, on_delete=models.CASCADE)
    number = models.CharField(max_length=100, verbose_name=_('Number'))
    notes = models.TextField(verbose_name=_('Notes'), blank=True)
    bundle = models.FileField(upload_to='applications', blank=True, null=True)
    creation_timestamp = models.DateTimeField(default=now)
    last_update_timestamp = models.DateTimeField(auto_now=True)
    deployment_status = JSONField(blank=True, default=dict)

    def __repr__(self):
        return '<Version %s>' % self.application.slug

    def create_bundle(self):
        app = self.application
        elements = app.scandeps()
        tar_io = io.BytesIO()
        with tarfile.open(mode='w', fileobj=tar_io) as tar:
            manifest_json = {
                'application': app.name,
                'slug': app.slug,
                'description': app.description,
                'documentation_url': app.documentation_url,
                'icon': os.path.basename(app.icon.name) if app.icon.name else None,
                'version_number': self.number,
                'version_notes': self.notes,
                'elements': [],
            }

            for element, relation in elements.values():
                manifest_json['elements'].append(
                    {
                        'type': element.type,
                        'slug': element.slug,
                        'name': element.name,
                        'auto-dependency': relation.auto_dependency,
                    }
                )

                response = requests.get(element.cache['urls']['export'])
                tarinfo = tarfile.TarInfo('%s/%s' % (element.type, element.slug))
                tarinfo.mtime = self.last_update_timestamp.timestamp()
                tarinfo.size = int(response.headers['content-length'])
                tar.addfile(tarinfo, fileobj=io.BytesIO(response.content))

            manifest_fd = io.BytesIO(json.dumps(manifest_json, indent=2).encode())
            tarinfo = tarfile.TarInfo('manifest.json')
            tarinfo.size = len(manifest_fd.getvalue())
            tarinfo.mtime = self.last_update_timestamp.timestamp()
            tar.addfile(tarinfo, fileobj=manifest_fd)

            if app.icon.name:
                icon_fd = app.icon.file
                tarinfo = tarfile.TarInfo(manifest_json['icon'])
                tarinfo.size = icon_fd.size
                tarinfo.mtime = self.last_update_timestamp.timestamp()
                tar.addfile(tarinfo, fileobj=icon_fd)

        self.bundle.save('%s.tar' % app.slug, content=ContentFile(tar_io.getvalue()))
        self.save()

    def deploy(self):
        bundle_content = self.bundle.read()
        self.deploy_roles(bundle_content)
        for service_id, services in getattr(settings, 'KNOWN_SERVICES', {}).items():
            if service_id not in Application.SUPPORTED_MODULES:
                continue
            service_objects = {x.get_base_url_path(): x for x in get_installed_services(types=[service_id])}
            for service in services.values():
                if service_objects[service['url']].secondary:
                    continue
                url = urllib.parse.urljoin(service['url'], 'api/export-import/bundle-import/')
                response = requests.put(url, data=bundle_content)
                if not response.ok:
                    # TODO: report failures
                    continue
                # TODO: look at response content for afterjob URLs to display a progress bar
                pass

    def get_authentic_service(self):
        for service_id, services in getattr(settings, 'KNOWN_SERVICES', {}).items():
            if service_id == 'authentic':
                for service in services.values():
                    return service
        return None

    def deploy_roles(self, bundle):
        tar_io = io.BytesIO(bundle)
        service = self.get_authentic_service()
        if not service:
            return
        roles_api_url = urllib.parse.urljoin(service['url'], 'api/roles/?update_or_create=slug')
        provision_api_url = urllib.parse.urljoin(service['url'], 'api/provision/')
        with tarfile.open(fileobj=tar_io) as tar:
            manifest = json.loads(tar.extractfile('manifest.json').read().decode())
            for element in manifest.get('elements'):
                if element.get('type') != 'roles':
                    continue
                role_info = json.loads(tar.extractfile('%s/%s' % (element['type'], element['slug'])).read())
                # create or update
                response = requests.post(roles_api_url, json=role_info)
                if not response.ok:
                    # TODO: report failures
                    continue
                # then force provisionning
                response = requests.post(provision_api_url, json={'role_uuid': response.json()['uuid']})
                if not response.ok:
                    # TODO: report failures
                    continue


class AsyncJob(models.Model):
    label = models.CharField(max_length=100)
    status = models.CharField(
        max_length=100,
        default='registered',
        choices=[
            ('registered', _('Registered')),
            ('running', _('Running')),
            ('failed', _('Failed')),
            ('completed', _('Completed')),
        ],
    )
    creation_timestamp = models.DateTimeField(default=now)
    last_update_timestamp = models.DateTimeField(auto_now=True)
    completion_timestamp = models.DateTimeField(default=None, null=True)
    exception = models.TextField()

    application = models.ForeignKey(Application, on_delete=models.CASCADE)
    version = models.ForeignKey(Version, on_delete=models.CASCADE, null=True)
    action = models.CharField(max_length=100)

    raise_exception = True

    def run(self, spool=False):
        if 'uwsgi' in sys.modules and spool:
            from hobo.applications.spooler import run_job

            tenant = getattr(connection, 'tenant', None)
            domain = getattr(tenant, 'domain_url', '')
            run_job.spool(domain=domain.encode(), job_id=str(self.pk).encode())
            return
        self.status = 'running'
        self.save()
        try:
            if self.action == 'scandeps':
                self.application.scandeps()
            elif self.action == 'create_bundle':
                self.version.create_bundle()
        except Exception as e:
            if self.raise_exception:
                raise
            self.status = 'failed'
            self.exception = traceback.format_exc()
        finally:
            self.status = 'completed'
        self.completion_timestamp = now()
        self.save()
