# hobo - portal to configure and deploy applications
# Copyright (C) 2015-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import json
import tarfile
import urllib.parse

from django.conf import settings
from django.core.files.base import ContentFile
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse, reverse_lazy
from django.utils.translation import ugettext_lazy as _
from django.views.generic import DetailView, FormView, ListView, TemplateView
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from hobo.environment.utils import get_installed_services

from .forms import GenerateForm, InstallForm, MetadataForm
from .models import Application, AsyncJob, Element, Relation, Version
from .utils import Requests

requests = Requests()


class HomeView(ListView):
    template_name = 'hobo/applications/home.html'
    model = Application

    def get_queryset(self):
        return super().get_queryset().order_by('name')


home = HomeView.as_view()


class InitView(CreateView):
    template_name = 'hobo/applications/create.html'
    model = Application
    fields = ['name']

    def get_success_url(self):
        return reverse('application-manifest', kwargs={'app_slug': self.object.slug})


init = InitView.as_view()


def get_object_types():
    object_types = []
    for service_id, services in getattr(settings, 'KNOWN_SERVICES', {}).items():
        if service_id not in Application.SUPPORTED_MODULES:
            continue
        service_objects = {x.get_base_url_path(): x for x in get_installed_services(types=[service_id])}
        for service in services.values():
            if service['url'] not in service_objects:
                continue
            if service_objects[service['url']].secondary:
                continue
            url = urllib.parse.urljoin(service['url'], 'api/export-import/')
            response = requests.get(url)
            if not response.ok:
                continue
            for object_type in response.json()['data']:
                object_type['service'] = service
                object_types.append(object_type)
    return object_types


class ManifestView(TemplateView):
    template_name = 'hobo/applications/manifest.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['app'] = Application.objects.get(slug=self.kwargs['app_slug'])

        context['relations'] = context['app'].relation_set.all().select_related('element')
        context['last_version'] = context['app'].version_set.order_by('last_update_timestamp').last()
        context['types_by_service'] = {}

        type_labels = {}
        for object_type in get_object_types():
            type_labels[object_type['id']] = object_type['singular']
            if object_type.get('minor'):
                continue
            service = object_type['service']['title']
            if service not in context['types_by_service']:
                context['types_by_service'][service] = []
            context['types_by_service'][service].append(object_type)

        for relation in context['relations']:
            relation.element.type_label = type_labels.get(relation.element.type)

        return context


manifest = ManifestView.as_view()


class VersionsView(ListView):
    template_name = 'hobo/applications/versions.html'

    def get_queryset(self):
        self.app = Application.objects.get(slug=self.kwargs['app_slug'])
        return self.app.version_set.order_by('-last_update_timestamp')

    def get_context_data(self, **kwargs):
        kwargs['app'] = self.app
        return super().get_context_data(**kwargs)


versions = VersionsView.as_view()


class MetadataView(UpdateView):
    template_name = 'hobo/applications/edit-metadata.html'
    model = Application
    slug_url_kwarg = 'app_slug'
    form_class = MetadataForm

    def get_success_url(self):
        return reverse('application-manifest', kwargs={'app_slug': self.object.slug})


metadata = MetadataView.as_view()


class AppAddElementView(TemplateView):
    template_name = 'hobo/applications/add-element.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['app'] = Application.objects.get(slug=self.kwargs['app_slug'])
        for object_type in get_object_types():
            if object_type.get('id') == self.kwargs['type']:
                context['type'] = object_type
                url = object_type['urls']['list']
                response = requests.get(url)
                context['elements'] = response.json()['data']
                break
        return context

    def post(self, request, app_slug, type):
        context = self.get_context_data()
        app = context['app']
        element_infos = {x['id']: x for x in context['elements']}
        for element_slug in request.POST.getlist('elements'):
            element, created = Element.objects.get_or_create(
                type=type, slug=element_slug, defaults={'name': element_infos[element_slug]['text']}
            )
            element.name = element_infos[element_slug]['text']
            element.cache = element_infos[element_slug]
            element.save()
            relation, created = Relation.objects.get_or_create(application=app, element=element)
            relation.auto_dependency = False
            relation.save()
        return HttpResponseRedirect(reverse('application-manifest', kwargs={'app_slug': app_slug}))


add_element = AppAddElementView.as_view()


class AppDeleteElementView(DeleteView):
    model = Relation
    template_name = 'hobo/applications/element_confirm_delete.html'

    def get_success_url(self):
        return reverse('application-manifest', kwargs={'app_slug': self.kwargs['app_slug']})


delete_element = AppDeleteElementView.as_view()


def scandeps(request, app_slug):
    job = AsyncJob(
        label=_('Scanning for dependencies'),
        application=Application.objects.get(slug=app_slug),
        action='scandeps',
    )
    job.save()
    job.run(spool=True)
    if job.status == 'registered':
        return HttpResponseRedirect(
            reverse('application-async-job', kwargs={'app_slug': app_slug, 'pk': job.id})
        )

    return HttpResponseRedirect(reverse('application-manifest', kwargs={'app_slug': app_slug}))


class GenerateView(FormView):
    form_class = GenerateForm
    template_name = 'hobo/applications/generate.html'

    def get_initial(self):
        self.app = Application.objects.get(slug=self.kwargs['app_slug'])
        version = self.app.version_set.order_by('last_update_timestamp').last()
        if version:
            self.initial['number'] = version.number
            self.initial['notes'] = version.notes
        return super().get_initial()

    def form_valid(self, form):
        app = self.app

        version = (
            app.version_set.filter(number=form.cleaned_data['number'])
            .order_by('last_update_timestamp')
            .last()
        )
        if not version:
            version = Version(application=app)
        version.number = form.cleaned_data['number']
        version.notes = form.cleaned_data['notes']
        version.save()

        job = AsyncJob(
            label=_('Creating application bundle'),
            application=app,
            version=version,
            action='create_bundle',
        )
        job.save()
        job.run(spool=True)
        if job.status == 'registered':
            return HttpResponseRedirect(
                reverse('application-async-job', kwargs={'app_slug': app.slug, 'pk': job.id})
            )

        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        kwargs['app_slug'] = self.kwargs['app_slug']
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return reverse('application-manifest', kwargs={'app_slug': self.kwargs['app_slug']})


generate = GenerateView.as_view()


def download(request, app_slug, version_pk=None):
    app = Application.objects.get(slug=app_slug)
    if version_pk is None:
        version = app.version_set.order_by('last_update_timestamp').last()
    else:
        version = get_object_or_404(app.version_set, pk=version_pk)
    response = HttpResponse(version.bundle, content_type='application/x-tar')
    response['Content-Disposition'] = 'attachment; filename="%s"' % '%s.tar' % app_slug
    return response


class Install(FormView):
    form_class = InstallForm
    template_name = 'hobo/applications/install.html'
    success_url = reverse_lazy('applications-home')

    def form_valid(self, form):
        tar_io = io.BytesIO(self.request.FILES['bundle'].read())
        tar = tarfile.open(fileobj=tar_io)
        manifest = json.loads(tar.extractfile('manifest.json').read().decode())
        app, created = Application.objects.get_or_create(
            slug=manifest.get('slug'), defaults={'name': manifest.get('application')}
        )
        app.name = manifest.get('application')
        app.description = manifest.get('description')
        app.documentation_url = manifest.get('documentation_url') or ''
        if created:
            # mark as non-editable only newly deployed applications, this allows
            # overwriting a local application and keep on developing it.
            app.editable = False
        else:
            app.relation_set.all().delete()
        app.save()
        icon = manifest.get('icon')
        if icon:
            app.icon.save(icon, tar.extractfile(icon), save=True)
        else:
            app.icon.delete()

        for element_dict in manifest.get('elements'):
            element, created = Element.objects.get_or_create(
                type=element_dict['type'], slug=element_dict['slug'], defaults={'name': element_dict['name']}
            )
            element.name = element_dict['name']
            element.save()

            relation = Relation(
                application=app,
                element=element,
                auto_dependency=element_dict['auto-dependency'],
            )
            relation.save()

        # always create a new version on install
        version = Version(
            application=app,
            number=manifest.get('version_number') or 'unknown',
            notes=manifest.get('version_notes') or '',
        )
        version.bundle.save('%s.tar' % app.slug, content=ContentFile(tar_io.getvalue()))
        version.save()

        version.deploy()

        return super().form_valid(form)


install = Install.as_view()


class AppDeleteView(DeleteView):
    model = Application
    template_name = 'hobo/applications/app_confirm_delete.html'

    def get_success_url(self):
        return reverse('applications-home')


delete = AppDeleteView.as_view()


class AsyncJobView(DetailView):
    model = AsyncJob
    template_name = 'hobo/applications/job.html'

    def get_redirect_url(self):
        return reverse('application-manifest', kwargs={'app_slug': self.kwargs['app_slug']})


async_job = AsyncJobView.as_view()
